#include "types.h"
#include "mrmidi2.h"
#include "mmc.h"
#ifndef LED_DISP
	#include "lcd.h"
#else
	#define lcd_data(a) ;
	#define lcd_setcur(a, b) ;
	#define lcd_data(a) ;
	#define lcd_string(a, b) ;
	#define lcd_init() ;
	#define lcd_hex_u08(a) ;
	#define lcd_hex_u16(a) ;
	#define lcd_number(a, b) ;
	#define lcd_number_k(a) ;
#endif
#include "delay.h"
#include "fat16.h"
#ifdef USE_RTC
	#include "clock.h"
#endif
#include <avr/pgmspace.h>

struct fat_filedata_st fat_filedata;
static u16 fat_pos, fat_size, file_pos, root_pos, sect_offs/*, totalcl,*/;
static u08 cluster_size, fat_cnt;
#ifndef NO_RECORD
	static u16 startfatoffs, startfatcnt, clcnt;
	static u08 last_rdoffs, maxfilenum=0;
#endif
#ifdef DIR_SUPPORT
#ifndef NO_RECORD
#ifdef WR_DIR_SUPPORT
	static u32 last_rdsect=0;
#else
	static u16 last_rdsect=0;
#endif
#endif
	u16 dircl=0;
	static u08 dirfatflag=0;
	#define DIR_ALLOCATE_CLUSTER 1
	#define NEW_CLUSTER_ALLOC 2
#else
#ifndef NO_RECORD
	static u16 last_rdsect=0;
#endif
#endif
u08 filemode=0, fat_directaccess=0;
#ifdef BUFFERED_FILEENTRIES
	#define BFE_STEP	5
	u16 gl_fenr;
	u16 fe_buf[(1<<(8-BFE_STEP))-1];
#endif

#ifdef USE_RTC
	struct time_st {
		u08 gl_sek, gl_min, gl_std, gl_tag, gl_mon, gl_jahr;
	};
	extern struct time_st gl_time;
#endif
extern u08 sw_version;
extern unsigned char sharedmem[];
extern u16 file_cnt, bufpos;
extern u32 sect;
#ifdef SD_TIMEOUT
	extern u08 state;
	#define ERR 0x80
#endif
#ifdef LFN_SCROLLING
	extern u08 lfnpos;
#endif
#if defined(PLAYLIST) || defined(LFN_SCROLLING)
	extern u08 lfnlen;
#endif

static u16up find_free_cluster(void);

#ifndef USE_RTC
prog_uchar FILEENTRY[32-11-4-2] = {
0x20, 0x00, 0xc2, 0xb3, 0x92, 0x84, 0x36, 0x84, 0x36, 0,0, 0x91, 0x92, 0x84, 0x36};
#endif
prog_uchar FILEDEFNAME[] = "MM      MID";

//Block Size in Bytes
#define BlockSize			512

//Master Boot Record
#define MASTER_BOOT_RECORD_SECTOR	0

//Volume Boot Record location in Master Boot Record
#define VBR_ADDR 			0x1C6

//define ASCII
#define SPACE 				0x20
#define DIR_ENTRY_IS_DELETED   0xE5
#define DIR_ENTRY_IS_DELETED2  0x05
#define FIRST_LONG_ENTRY	0x01
#define SECOND_LONG_ENTRY	0x42

//define DIR_Attr
#define ATTR_LONG_NAME		0x0F
#define ATTR_READ_ONLY		0x01
#define ATTR_HIDDEN			0x02
#define ATTR_SYSTEM			0x04
#define ATTR_VOLUME_ID		0x08
#define ATTR_DIRECTORY		0x10
#define ATTR_ARCHIVE		0x20

#ifdef DIR_SUPPORT
	#define DIR_UP_FILE	 0
#endif

//! Usage requires little endian compiler!
struct BootSec_st
{
	u08 BS_jmpBoot[3];
	u08 BS_OEMName[8];
	u16 BPB_BytesPerSec; //2 bytes
	u08 BPB_SecPerClus;
	u16 BPB_RsvdSecCnt; //2 bytes
	u08 BPB_NumFATs;
	u16 BPB_RootEntCnt; //2 bytes
	u16 BPB_TotSec16; //2 bytes
	u08 BPB_Media;
	u16 BPB_FATSz16; //2 bytes
	u16 BPB_SecPerTrk; //2 bytes
	u16 BPB_NumHeads; //2 bytes
	u32 BPB_HiddSec; //4 bytes
	u32 BPB_TotSec32; //4 bytes
};

//FAT12 and FAT16 Structure Starting at Offset 36
#define BS_DRVNUM			36
#define BS_RESERVED1		37
#define BS_BOOTSIG			38
#define BS_VOLID			39
#define BS_VOLLAB			43
#define BS_FILSYSTYPE		54

//FAT32 Structure Starting at Offset 36
#define BPB_FATSZ32			36
#define BPB_EXTFLAGS		40
#define BPB_FSVER			42
#define BPB_ROOTCLUS		44
#define BPB_FSINFO			48
#define BPB_BKBOOTSEC		50
#define BPB_RESERVED		52

#define FAT32_BS_DRVNUM		64
#define FAT32_BS_RESERVED1	65
#define FAT32_BS_BOOTSIG	66
#define FAT32_BS_VOLID		67
#define FAT32_BS_VOLLAB		71
#define FAT32_BS_FILSYSTYPE	82
//End of Boot Sctor and BPB Structure

//! Usage requires little endian compiler!
struct DirEntry_st {
	u08	DIR_Name[11];	 //8 chars filename
	u08	DIR_Attr;		 //file attributes RSHA, Longname, Drive Label, Directory
	u08	DIR_NTRes;		//set to zero
	u08	DIR_CrtTimeTenth; //creation time part in milliseconds
	u16	DIR_CrtTime;	  //creation time
	u16	DIR_CrtDate;	  //creation date
	u16	DIR_LastAccDate;  //last access date
	u16	DIR_FstClusHI;	//first cluster high word
	u16	DIR_WrtTime;	  //last write time
	u16	DIR_WrtDate;	  //last write date
	u16	DIR_FstClusLO;	//first cluster low word
	u32	DIR_FileSize;
};


#ifdef USE_RTC

/** Generiert Format der FAT-Zeit.
*/
inline static u16up fat_maketime(void)
{
	return (gl_time.gl_sek>>1) | ((u16up)gl_time.gl_min<<5) | ((u16up)gl_time.gl_std<<11);
}


/** Generiert Format des FAT-Datums.
*/
inline static u16up fat_makedate(void)
{
	return (gl_time.gl_tag) | ((u16up)gl_time.gl_mon<<5) | ((u16up)(gl_time.gl_jahr+20)<<9);
}
#endif


/** Calc Cluster from given Sector
	@param Sector
	@return Cluster
*/
static u16up sector2cluster(u32up s) {
	u08up cs = cluster_size;
	u32up temp = s - file_pos;

	while (!(cs & 1)) {
		cs >>= 1;
		temp >>= 1;
	}
	return temp+2;
}


/** Calc Sector from given Cluster
	@param Cluster
	@return Sector
*/
u32up cluster2sector(u16up c) {
	u08up cs = cluster_size;
	u32up s;

	s = c - 2;
	while (!(cs & 1)) {
		cs >>= 1;
		s <<= 1;
	}
	return s + file_pos;
}


/** Print Filename on Display
	@param line number [0,1]
*/
void print_filename(u08up y) {
#ifndef LED_DISP
	u08up i;
#ifndef LFN_SCROLLING
	u08 *ptr;
#endif
#ifdef NO_FILE_EXT
	u08up p;
#endif
	lcd_setcur(0, y);
#ifndef LFN_SCROLLING
	ptr = fat_filedata.name;
#endif
#ifdef NO_FILE_EXT
	p = NAMESIZE;
	while (--p && fat_filedata.name[p] != '.');
	if (p == 0)
		p = NAMESIZE;
	for (i=0; i<p; i++)
#else			
	for (i=0; i<LINESIZE; i++)
#endif
#ifndef LFN_SCROLLING
		lcd_data(*ptr++);
#else
		if (i+lfnpos < NAMESIZE)
			lcd_data(fat_filedata.name[i+lfnpos]);
#endif
#ifdef NO_FILE_EXT
	for (i=p; i<NAMESIZE; i++)
		lcd_data(' ');
#endif
#ifdef DIR_SUPPORT
	if (fat_filedata.attr & ATTR_DIRECTORY) {
		lcd_setcur(DIR_POS_X, y);
		lcd_data('<');
#ifdef COMPACT16x2
		lcd_data('D');
#else
		lcd_data('D');lcd_data('I');lcd_data('R');
#endif
		lcd_data('>');
	}
#endif
#endif
}


/** Print Filesize on Display
*/
void print_filesize(void) {
#ifndef LED_DISP
	lcd_setcur(FSIZE_X, FSIZE_Y);
	if (fat_filedata.len < 10000)
		lcd_number_k(fat_filedata.len);
	else {
		lcd_number(fat_filedata.len>>10, 1);
		lcd_data('k');
	}
#endif
}


/** Read Long File Name
	Reads data from sharedmem
	Stores result in global struct fat_filedata.name
*/
static void extractlfn(void) {
	u08up i, spos = 1, pos;

	pos = ((sharedmem[0]&0x3f)-1)*13;

	for (i=0; i<13; i++) {
		if (!sharedmem[spos] || sharedmem[spos] == 0xff)
			return;
		if (pos < NAMESIZE)
		{
#ifdef JAPAN_KANA
			u08up temp, c = sharedmem[spos+1];
			if (c == 0xff)
				temp = sharedmem[spos] + 0x40;
#ifdef JAPAN_KANA_EXTENDED
			else if (c == 0x30)
				temp = KANA[sharedmem[spos]]; // TODO
#endif
			else
				temp = sharedmem[spos];
			fat_filedata.name[pos++] = temp;
#else
			fat_filedata.name[pos++] = sharedmem[spos];
#endif
#if defined(PLAYLIST) || defined(LFN_SCROLLING)
			if (pos > lfnlen)
				lfnlen = pos;
#endif
		}
		else
			return;
		spos += 2;
		if (spos == 11)
			spos = 14;
		else if (spos == 26)
			spos = 28;
	}
}


/** Filesystem-Init
	@return success [=true]
*/
u08up fat_init(void) {
	u16up root_size;
	struct BootSec_st *bs;
	u32up data_sec, cl;

	mmc_read_sector(MASTER_BOOT_RECORD_SECTOR, sharedmem); // MBR lesen
	sect_offs = sharedmem[VBR_ADDR] | (sharedmem[VBR_ADDR+1]<<8); // Erste Partition: Startsektor bestimmen
	if (sect_offs > 1000) // Ungltiger Startsektor - TODO check
		sect_offs = 0;  // Kein MBR? Annahme: SuperFloppy-Format (ohne Partitionen)

	mmc_read_sector(sect_offs, sharedmem); // Bootsektor lesen
#ifdef SD_TIMEOUT
	if (state >= ERR)
		return 0;
#endif
	bs = (struct BootSec_st *)sharedmem; // Strukturinterpretation anbringen
	fat_size = bs->BPB_FATSz16; // FAT-Gre auslesen
	if (fat_size == 0) // Ist nicht angegeben... wohl FAT32 oder ganz was Anderes
		return 0;

	cluster_size = bs->BPB_SecPerClus; // Sektoren pro Cluster auslesen
	fat_cnt = bs->BPB_NumFATs; // Anzahl FATs lesen
	if (fat_cnt == 0) // FAT-Anzahl ungltig
		return 0;

	fat_pos = sect_offs + bs->BPB_RsvdSecCnt; // Reservierte Sektoren aufaddieren -> FAT-Startsektor //(sharedmem[0x0e] | (sharedmem[0x0f]<<8)); // alt
	root_pos = fat_pos + fat_size * fat_cnt; // Rootdirectory-Startsektor berechnen
	root_size = bs->BPB_RootEntCnt; // Anzahl reservierte Rootdir-Eintrge //sharedmem[0x11] | (sharedmem[0x12]<<8); // alt
	if (root_size == 0) // Ungltige Anzahl
		return 0;
	file_pos = root_pos + ((root_size+15)>>4); // Erster Datensektor - TODO: verify! - Anm.: 512 Bytes/Sector hardwired
	if (bs->BPB_TotSec16 != 0)
		data_sec = bs->BPB_TotSec16;
	else
		data_sec = bs->BPB_TotSec32;
	data_sec = data_sec - file_pos;
	cl = data_sec / cluster_size;
	if (cl < 4085) {
		/* Volume is FAT12 */
		return 0;
	}
	else if (cl < 65525) {
		//totalcl = (u16)cl;
		/* Volume is FAT16 */
	}
	else {
		return 0;
		/* Volume is FAT32 */
	}

	return 1;
}


#ifndef NO_RECORD

/** Char to Number Conversion, always [3 chars] to [byte]
	@param String
	@return Number
*/
static u08up atoi(unsigned char *s) {
	return (s[0]-'0')*100+(s[1]-'0')*10+(s[2]-'0');
}
#endif


/** Fileentry-Reader
	Reads sequential MMC data, must be set up to start sector of root dir
	Stores information in global struct fat_filedata
	@return success [=true]
*/
static u08up read_fileentry(void) {
	struct DirEntry_st *de;
	u16up dirsect;
	u08up lfn=0, i, c=0, e=0, dirsectoffs, fnum, *p;

	p = fat_filedata.name;
	for (i=0; i<NAMESIZE; i++)
		*p++ = ' ';
	fat_filedata.dirsect = 0;
#if defined(PLAYLIST) || defined(LFN_SCROLLING)
	lfnlen = 0;
#endif
	do {
		dirsect = sect;
		dirsectoffs = bufpos>>5;
#ifdef DIR_SUPPORT
		dirfatflag = 0;
#endif
		mmc_read_block(sharedmem, 32);
#ifdef BUFFERED_FILEENTRIES
		gl_fenr++;
#endif
#ifdef SD_TIMEOUT
		if (state >= ERR)
			return 0;
#endif
		de = (struct DirEntry_st *)sharedmem;
		if (
#ifdef DIR_SUPPORT
			dircl == 0 &&
#endif
		sect >= file_pos) {
#ifndef NO_RECORD
			last_rdsect = 0; // Kein Platz mehr im Rootdir!
#endif
			return 0;
		}
		else if (sharedmem[0] == 0
#ifdef DIR_SUPPORT
			|| (dircl && (dirfatflag & NEW_CLUSTER_ALLOC))//bufpos == 32 && sect == dirsect) // wollte vergeblich neuen Cluster angefangen
#endif
		)
		{
#ifndef NO_RECORD
			if (last_rdsect == 0) {
				last_rdsect = dirsect;
				last_rdoffs = dirsectoffs;
			}
#endif
			//continue;
			return 0;
		}
		else if (sharedmem[0] == DIR_ENTRY_IS_DELETED || sharedmem[0] == DIR_ENTRY_IS_DELETED2 || de->DIR_Attr == ATTR_VOLUME_ID) // Gelscht oder Volumebezeichner
			continue;
		else if (de->DIR_Attr == ATTR_LONG_NAME) {
			if (fat_filedata.dirsect == 0) {
				fat_filedata.dirsect = dirsect;
				fat_filedata.dirsectoffs = dirsectoffs;
			}
			extractlfn();
			lfn = 1;
			e++;
		}
		else {
			fat_filedata.entries = e+1;
			if (!lfn) {
#if defined(PLAYLIST) || defined(LFN_SCROLLING)
				lfnlen = 8+4;
#endif
				fat_filedata.dirsect = dirsect;
				fat_filedata.dirsectoffs = dirsectoffs;
				for (i=0; i<8; i++)
					fat_filedata.name[c++] = sharedmem[i]; // de->DIR_Name[i]
#ifdef DIR_SUPPORT
				if (de->DIR_Attr & ATTR_DIRECTORY)
					fat_filedata.name[c++] = ' ';
				else
#endif
					fat_filedata.name[c++] = '.';
				for (i=8; i<11; i++)
					fat_filedata.name[c++] = sharedmem[i];
#ifndef NO_RECORD
				if (fat_filedata.name[0] == 'M' && fat_filedata.name[1] == 'M'
				  && fat_filedata.name[2] >= '0' && fat_filedata.name[2] <= '2'
				  && fat_filedata.name[3] >= '0' && fat_filedata.name[3] <= '9'
				  && fat_filedata.name[4] >= '0' && fat_filedata.name[4] <= '9') {
				  	fnum = atoi(&fat_filedata.name[2]);
				  	if (fnum > maxfilenum)
						maxfilenum = fnum;
				}
#endif
			}
			fat_filedata.len = de->DIR_FileSize; //sharedmem[0x1c] | (sharedmem[0x1d]<<8) | ((u32)sharedmem[0x1e]<<16) | ((u32)sharedmem[0x1f]<<24);
			fat_filedata.startcl = de->DIR_FstClusLO; //cluster2sector(sharedmem[0x1a] | (sharedmem[0x1b]<<8));
#ifdef DIR_SUPPORT
			fat_filedata.attr = de->DIR_Attr;
			if (fat_filedata.name[0] == '.' && fat_filedata.name[1] == ' ')
			{
				lfn = 0; c = 0; e = 0; // War nur dot, also weitersuchen
#if defined(PLAYLIST) || defined(LFN_SCROLLING)
				lfnlen = 0;
#endif
				fat_filedata.dirsect = 0;
			}
#else
			if (de->DIR_Attr & ATTR_DIRECTORY)
			{
				lfn = 0; c = 0; e = 0; // War keine Datei, also weitersuchen
#if defined(PLAYLIST) || defined(LFN_SCROLLING)
				lfnlen = 0;
#endif
				fat_filedata.dirsect = 0;
			}
#endif
			else {
				return 1;
			}
		}
	} while (1);
}


/** Searches for a file entry with a given index
	@param File Number
	@return File found [=true]
*/
u08up fat_read_filedata(u08up nr) {	// new file data in global struct
#ifdef BUFFERED_FILEENTRIES
	u32up sector;
	u16up fenr;
	u08up bnr = nr>>BFE_STEP;
#endif
	if (nr >= file_cnt)
		return 0;
#ifdef DIR_SUPPORT
	if (!dircl) // Rootdir
#endif
		fat_directaccess = 1;
#ifdef BUFFERED_FILEENTRIES
	fenr = fe_buf[bnr];
	if (!dircl) // Rootdir
		sector = rootpos;
	else
		sector = cluster2sector(dircl);
	mmc_load_start(sector + (fenr>>4));
	fenr &= 0xf;
	fenr <<= 5;
	while (fenr--)
		mmc_fetch_byte();
	nr &= (1<<BFE_STEP)-1;
#else
#ifdef DIR_SUPPORT
	if (!dircl) // Rootdir
		mmc_load_start(root_pos);
	else
		mmc_load_start(cluster2sector(dircl));
#else
	mmc_load_start(root_pos);
#endif
#endif
#ifdef SD_TIMEOUT
	if (state >= ERR)
		return 0;
#endif
	if (!read_fileentry()) {
		lcd_string(DISP_CARDERR, LINE_CARDERR);
		delay_ms(1000);
		fat_directaccess = 0;
		return 0;
	}
	while (nr--)
		if (!read_fileentry()) {
			lcd_string(DISP_CARDERR, LINE_CARDERR);
			delay_ms(1000);
			fat_directaccess = 0;
			return 0;
		}
	mmc_complete_read();
#ifdef DIR_SUPPORT
	if (!dircl) // Rootdir
#endif
		fat_directaccess = 0;
	return 1;
}


/** Count Files in Dir.
	Stores result in global var file_cnt.
*/
void fat_count_files(void) {
#ifdef BUFFERED_FILEENTRIES
	u08up pp=0, cp=0;
	gl_fenr = 0;
#endif
#ifdef DIR_SUPPORT
	if (!dircl) // Rootdir
#endif
		fat_directaccess = 1;
#ifdef DIR_SUPPORT
	if (!dircl) // Rootdir
		mmc_load_start(root_pos);
	else {
		mmc_load_start(cluster2sector(dircl));
	}
#else
	mmc_load_start(root_pos);
#endif
#ifdef SD_TIMEOUT
	if (state >= ERR)
		return 0;
#endif
	file_cnt = 0;
#ifndef NO_RECORD
	last_rdsect = 0;
#endif
#ifdef DIR_SUPPORT
	dirfatflag = 0;
#endif
	while (read_fileentry()) {
#ifdef BUFFERED_FILEENTRIES
		pp++;
		if (pp == (1<<BFE_STEP)) {
			pp = 0;
			fe_buf[cp++] = gl_fenr;
		}
#endif
		if (file_cnt == 0
#ifdef DIR_SUPPORT
			&& dircl == 0 // FW muss im Rootdir sein!
#endif
			)
		{
#if defined(OLD_BL) || defined(NEW_BL)
			if (fat_filedata.name[0] == 'M' && fat_filedata.name[1] == 'M' && fat_filedata.name[2] == ' '
			  && fat_filedata.name[9] == 'B' && fat_filedata.name[10] == 'I' && fat_filedata.name[11] == 'N') {
				mmc_complete_read();
				cli();
				TCNT1 = cluster2sector(fat_filedata.startcl);
				OCR1A = ((fat_filedata.len-1)>>9)+1;
				if (sw_version & 0x80) {
#ifdef __AVR_ATmega168__
					asm volatile("call 0x3E68");
				}
				else {
					if (OCR1A <= ((2*8192-2048-1)/512+1))
						asm volatile("call 0x3902");
#endif
#ifdef __AVR_ATmega88__
					asm volatile("call 0x1E68");
				}
				else {
					if (OCR1A <= ((8192-2048-1)/512+1))
						asm volatile("call 0x1902");
#endif
#ifdef __AVR_ATmega8__
					asm volatile("rcall 0x1E68");
#endif
				}
				//while (1);
			}
#endif
		}
		file_cnt++;
#ifndef DIR_SUPPORT
		lcd_setcur(BOOTFNUM_X, BOOTFNUM_Y);
		lcd_number(file_cnt, 1);	// Anzeige Anzahl gefundene Dateien in Echtzeit
#endif
	}
	mmc_complete_read();
#ifdef DIR_SUPPORT
	if (!dircl) // Rootdir
#endif
		fat_directaccess = 0;
#ifndef NO_RECORD
	if (last_rdoffs == 16) {
		last_rdoffs = 0;
#ifdef WR_DIR_SUPPORT
		if (dircl) {
			if (dirfatflag & NEW_CLUSTER_ALLOC) {
				dirfatflag = DIR_ALLOCATE_CLUSTER;
			}
			else {
				sect = last_rdsect;
				fat_getnextsector_rd();
				last_rdsect = sect;
			}
		}
		else
#endif
			last_rdsect++;
	}
#endif
}


#ifndef NO_RECORD
/** Deletes current file
	Removes FAT entries as well
*/
void fat_delete_file(void) {
#ifdef WR_DIR_SUPPORT
	u32 dirsect;
#endif
	u16up startcl = fat_filedata.startcl;	// Rechnet den Cluster der zu lschenden Datei aus
	u16up fatsect, fatcl, nfatsect;
	u08up offscl, i;

	// FATs bearbeiten
	for (i=0; i<fat_cnt; i++) {
		fatsect = fat_pos + (startcl>>8) + fat_size*i;	// Sektor der FAT mit dem Startcluster der zlD
		offscl = (u08)startcl;	// Eintragsposition im FAT-Sektor bestimmen
		mmc_read_sector(fatsect, sharedmem);	// FAT-Sektor lesen
		while (1) {
			fatcl = sharedmem[offscl<<1];	// FAT-Eintrag lesen
			fatcl |= sharedmem[(offscl<<1)+1] << 8;
			sharedmem[offscl<<1] = 0;	// Eintrag lschen (frei markieren)
			sharedmem[(offscl<<1)+1] = 0;
			if (fatcl < 0xfff0 && fatcl) {	// Wenn Eintrag auf weiteren FAT-Eintrag zeigt
				nfatsect = fat_pos + (fatcl>>8);	// Neuen FAT-Sektor bestimmen
				if (i == 1)
			 		nfatsect += fat_size;	// Kopie der FAT
				offscl = (u08)fatcl;	// Eintragsposition im FAT-Sektor bestimmen
				if (nfatsect != fatsect) {	// Falls nchster Eintrag in einem anderen FAT-Sektor liegt
					mmc_write_sector(fatsect, sharedmem);	// Genderten FAT-Sektor wegschreiben
#ifdef SD_TIMEOUT
					if (state >= ERR)
						return;
#endif
					fatsect = nfatsect;	// Neuen Sektor festlegen
					mmc_read_sector(fatsect, sharedmem);	// Diesen FAT-Sektor auslesen
				}
			}
			else {
				break;	// Letzter FAT-Eintrag gewesen: Schleife verlassen
			}
		}
		mmc_write_sector(fatsect, sharedmem);	// genderten FAT-Sektor wegschreiben
	}


#ifndef WR_DIR_SUPPORT
	// Direntry ndern
	fatsect = fat_filedata.dirsect;
	mmc_read_sector(fatsect, sharedmem);	// Betroffenen Rootdirsektor laden
	offscl = fat_filedata.dirsectoffs;

	for (i=0; i<fat_filedata.entries; i++) {	// Fr jeden Direintrag dieser Datei tue:
		if (offscl >= 16) {	// Sektorgrenze berschritten
			mmc_write_sector(fatsect, sharedmem);	// Dirsektor speichern
#ifdef SD_TIMEOUT
			if (state >= ERR)
				return;
#endif
			fatsect++;	// Nchsten Dirsektor bestimmen
			offscl = 0;
			mmc_read_sector(fatsect, sharedmem);	// Betroffenen Dirsektor laden
		}
		sharedmem[offscl*32] = DIR_ENTRY_IS_DELETED;	// Erstes Byte im Dateinamen auf "gelscht" setzen
		offscl++;
	}
	mmc_write_sector(fatsect, sharedmem);
#else
	// Direntry ndern
	dirsect = fat_filedata.dirsect;
	mmc_read_sector(dirsect, sharedmem);	// Betroffenen Rootdirsektor laden
	offscl = fat_filedata.dirsectoffs;

	for (i=0; i<fat_filedata.entries; i++) {	// Fr jeden Direintrag dieser Datei tue:
		if (offscl >= 16) { // Sektorgrenze berschritten
			mmc_write_sector(dirsect, sharedmem);   // Dirsektor speichern
#ifdef SD_TIMEOUT
			if (state >= ERR)
				return;
#endif
			if (dircl) {
				sect = dirsect;
				fat_getnextsector_rd();
				dirsect = sect;
			}
			else
				dirsect++;   // Nchsten Dirsektor bestimmen
			offscl = 0;
			mmc_read_sector(dirsect, sharedmem);	// Betroffenen Dirsektor laden
		}
		sharedmem[offscl*32] = DIR_ENTRY_IS_DELETED;	// Erstes Byte im Dateinamen auf "gelscht" setzen
		offscl++;
	}
	mmc_write_sector(dirsect, sharedmem);
#endif
}


/** Convert Number to String
	@param (OUT) String
	@param (IN) Number
*/
static void wrstr(u08 *s, u08up d) {
	u08up zehner=0, hunderter=0;
	while (d >= 100) {
		d -= 100;
		hunderter++;
	}
	while (d >= 10) {
		d -= 10;
		zehner++;
	}
	*s = hunderter+'0'; s++;
	*s = zehner+'0'; s++;
	*s = d+'0';
}


/** Create New File
	Uses information from global struct fat_filedata
*/
static u08up fat_new_file(void) {
#ifdef WR_DIR_SUPPORT
	u32up dirsect;
#endif
	u16up startcl = fat_filedata.startcl;	// Rechnet den Cluster der zu erstellenden Datei aus
	u16up fatsect, fatcl, lastfs, cnt;
	u08up offscl, i, lastco, dirty;
	u08 *p;
#ifdef USE_RTC
	struct DirEntry_st *modentry;
#endif

	if (last_rdsect == 0)
		return 0;

	for (i=0; i<12; i++)
		fat_filedata.name[i] = (u08)pgm_read_byte( &(FILEDEFNAME[i]) );
	wrstr(&(fat_filedata.name[2]), ++maxfilenum);

#ifndef WR_DIR_SUPPORT
	// Direintrag erstellen
	fat_filedata.entries = 1;
	fat_filedata.dirsect = last_rdsect;
	fat_filedata.dirsectoffs = last_rdoffs;
	fatsect = last_rdsect;
	mmc_read_sector(fatsect, sharedmem);	// Betroffenen Dirsektor laden
#ifdef SD_TIMEOUT
	if (state >= ERR)
		return;
#endif
	offscl = last_rdoffs;
	p = fat_filedata.name;
#ifdef USE_RTC
	modentry = (struct DirEntry_st *)sharedmem[offscl*32];
	for (i=0; i<11; i++)
		modentry->DIR_Name[i] = *p++;
	modentry->DIR_Attr = ATTR_ARCHIVE;
	modentry->DIR_NTRes = 0;
//#ifdef USE_RTC
	read_time();
	i = 0;
	if (gl_time.gl_sek&1)
		i = 100;
	modentry->DIR_CrtTimeTenth = i;
	modentry->DIR_CrtTime = fat_maketime();
	modentry->DIR_CrtDate = fat_makedate();
/*#else
	modentry->DIR_CrtTimeTenth = 0;
	modentry->DIR_CrtTime = (0) | (0<<5) | (12<<11);
	modentry->DIR_CrtDate = (5) | (9<<5) | (7<<9);
#endif*/
	modentry->DIR_LastAccDate = modentry->DIR_CrtDate;
	modentry->DIR_FstClusHI = 0;
	modentry->DIR_WrtTime = modentry->DIR_CrtTime;
	modentry->DIR_WrtDate = modentry->DIR_CrtDate;
	modentry->DIR_FstClusLO = startcl;
	modentry->DIR_FileSize = fat_filedata.len;
#else
	for (i=0; i<11; i++)
		sharedmem[offscl*32+i] = *p++;
	for (i=0; i<15; i++)
		sharedmem[offscl*32+i+11] = (u08)pgm_read_byte( &(FILEENTRY[i]) );
	sharedmem[offscl*32+31] = (u08)(fat_filedata.len >> 24);
	sharedmem[offscl*32+30] = (u08)(fat_filedata.len >> 16);
	sharedmem[offscl*32+29] = (u08)(fat_filedata.len >> 8);
	sharedmem[offscl*32+28] = (u08)fat_filedata.len;
	sharedmem[offscl*32+27] = (u08)(startcl >> 8);
	sharedmem[offscl*32+26] = (u08)startcl;
#endif
	mmc_write_sector(fatsect, sharedmem);
	
#else
	
	// Direintrag erstellen
	// Neuen Cluster in FAT-Eintragskette vom Parent-Dir erzeugen, falls erforderlich!
	if (dirfatflag & DIR_ALLOCATE_CLUSTER) {
		dirfatflag = 0;
		fatcl = find_free_cluster();

		cnt = sector2cluster(last_rdsect);
		fatsect = fat_pos + (cnt>>8);	// Sektor der FAT mit dem Cluster der Datei
		offscl = (u08)cnt;	// Eintragsposition im FAT-Sektor bestimmen
		mmc_read_sector(fatsect, sharedmem);	// FAT-Sektor lesen
		sharedmem[offscl<<1] = fatcl;	// FAT-Eintrag schreiben
		sharedmem[(offscl<<1)+1] = fatcl >> 8;
		mmc_write_sector(fatsect, sharedmem);	// FAT-Sektor schreiben

		fatsect = fat_pos + (fatcl>>8);	// Sektor der FAT mit dem Cluster der Datei
		offscl = (u08)fatcl;	// Eintragsposition im FAT-Sektor bestimmen
		mmc_read_sector(fatsect, sharedmem);	// FAT-Sektor lesen
		sharedmem[offscl<<1] = 0xff;	// FAT-Eintrag schreiben
		sharedmem[(offscl<<1)+1] = 0xff;
		mmc_write_sector(fatsect, sharedmem);	// FAT-Sektor schreiben

		//fat_write_val(sharedmem, sector2cluster(last_rdsect), fatcl);
		//fat_write_val(sharedmem, fatcl, 0xffff);
		last_rdsect = cluster2sector(fatcl);
	}

	fat_filedata.entries = 1;
	fat_filedata.dirsect = last_rdsect;
	fat_filedata.dirsectoffs = last_rdoffs;
		
	dirsect = last_rdsect;
	mmc_read_sector(dirsect, sharedmem);	// Betroffenen Dirsektor laden
#ifdef SD_TIMEOUT
	if (state >= ERR)
		return;
#endif
	offscl = last_rdoffs;
	p = fat_filedata.name;
#ifdef USE_RTC
	modentry = (struct DirEntry_st *)sharedmem[offscl*32];
	for (i=0; i<11; i++)
		modentry->DIR_Name[i] = *p++;
	modentry->DIR_Attr = ATTR_ARCHIVE;
	modentry->DIR_NTRes = 0;
//#ifdef USE_RTC
	read_time();
	i = 0;
	if (gl_time.gl_sek&1)
		i = 100;
	modentry->DIR_CrtTimeTenth = i;
	modentry->DIR_CrtTime = fat_maketime();
	modentry->DIR_CrtDate = fat_makedate();
/*#else
	modentry->DIR_CrtTimeTenth = 0;
	modentry->DIR_CrtTime = (0) | (0<<5) | (12<<11);
	modentry->DIR_CrtDate = (5) | (9<<5) | (7<<9);
#endif*/
	modentry->DIR_LastAccDate = modentry->DIR_CrtDate;
	modentry->DIR_FstClusHI = 0;
	modentry->DIR_WrtTime = modentry->DIR_CrtTime;
	modentry->DIR_WrtDate = modentry->DIR_CrtDate;
	modentry->DIR_FstClusLO = startcl;
	modentry->DIR_FileSize = fat_filedata.len;
#else
	for (i=0; i<11; i++)
		sharedmem[offscl*32+i] = *p++;
	for (i=0; i<15; i++)
		sharedmem[offscl*32+i+11] = (u08)pgm_read_byte( &(FILEENTRY[i]) );
	sharedmem[offscl*32+31] = (u08)(fat_filedata.len >> 24);
	sharedmem[offscl*32+30] = (u08)(fat_filedata.len >> 16);
	sharedmem[offscl*32+29] = (u08)(fat_filedata.len >> 8);
	sharedmem[offscl*32+28] = (u08)fat_filedata.len;
	sharedmem[offscl*32+27] = (u08)(startcl >> 8);
	sharedmem[offscl*32+26] = (u08)startcl;
#endif
	mmc_write_sector(dirsect, sharedmem);
#endif
	
	last_rdoffs++;
	if (last_rdoffs == 16) {
		last_rdoffs = 0;
		last_rdsect++;
	}

	// FATs bearbeiten
	for (i=0; i<fat_cnt; i++) {
		startcl = fat_filedata.startcl;	// Rechnet den Cluster der zu erstellenden Datei aus
		fatsect = fat_pos + (startcl>>8) + fat_size*i;	// Sektor der FAT mit dem Startcluster der zlD
		offscl = (u08)startcl;	// Eintragsposition im FAT-Sektor bestimmen
		cnt = clcnt;

		mmc_read_sector(fatsect, sharedmem);	// FAT-Sektor lesen
		if (cnt == 1) {	// Datei braucht blo 1 Cluster
			sharedmem[offscl<<1] = 0xff;	// FAT-Eintrag schreiben
			sharedmem[(offscl<<1)+1] = 0xff;
		}
		else {
			lastfs = fatsect;
			lastco = offscl;
			dirty = 0;
			while (1) {
				offscl++;
				//offscl &= 0xff;
				if (offscl == 0) {
					if (dirty) {
						mmc_write_sector(fatsect, sharedmem);
#ifdef SD_TIMEOUT
						if (state >= ERR)
							return;
#endif
						dirty = 0;
					}
					fatsect++;
					mmc_read_sector(fatsect, sharedmem);
				}
				fatcl = sharedmem[offscl<<1];	// FAT-Eintrag lesen
				fatcl |= sharedmem[(offscl<<1)+1] << 8;
				if (fatcl == 0) {
					if (lastfs != fatsect) {
						if (dirty) {
							mmc_write_sector(fatsect, sharedmem);
							dirty = 0;
						}
						mmc_read_sector(lastfs, sharedmem);
					}
					if (cnt > 1) {
						sharedmem[lastco<<1] = offscl;	// FAT-Eintrag schreiben
						sharedmem[(lastco<<1)+1] = fatsect - fat_pos - i*fat_size;
						cnt--;
					}
					else {
						sharedmem[lastco<<1] = 0xff;	// FAT-Eintrag schreiben
						sharedmem[(lastco<<1)+1] = 0xff;
						break;
					}
					dirty = 1;
					if (lastfs != fatsect) {
						mmc_write_sector(lastfs, sharedmem);
						dirty = 0;
						mmc_read_sector(fatsect, sharedmem);
					}
					lastfs = fatsect;
					lastco = offscl;
				}
			}
		}
		mmc_write_sector(fatsect, sharedmem);	// genderten FAT-Sektor wegschreiben
	}
	return 1;
}


/** Find next free cluster
	@return Cluster Number
*/
static u16up find_free_cluster(void) {
	u16up fatsect, fatcl, cnt=clcnt-startfatcnt, pcnt=startfatcnt;
	u08up offscl=0;

	// FAT bearbeiten
	fatsect = fat_pos+startfatoffs;	// Sektor der FAT
	if (startfatoffs == 0)
		offscl = 2;
	mmc_read_sector(fatsect, sharedmem);	// FAT-Sektor lesen
	while (1) {
		fatcl = sharedmem[offscl<<1];	// FAT-Eintrag lesen
		fatcl |= sharedmem[(offscl<<1)+1] << 8;
		if (fatcl == 0) {
			pcnt++;
			cnt--;
			if (cnt == 0)
				return ((fatsect-fat_pos)<<8) + offscl;
		}
		offscl++;
		//offscl &= 0xff;
		if (offscl == 0) {
			fatsect++;
			startfatoffs++;
			startfatcnt = pcnt;
//			if (((fatsect-fat_pos)<<8)+offscl == totalcl)	// Volle Karte! - TODO: verify!
//				return -1;
			mmc_read_sector(fatsect, sharedmem);	// FAT-Sektor lesen
#ifdef SD_TIMEOUT
			if (state >= ERR)
				return -1;
#endif
		}
	}
}
#endif


/** Opens a File
	@param mode
	@param file number (not needed for write operation)
	@return success
*/
u08up fat_openfile(u08up mode, u08up nr)
{
	if (mode == FILEMODE_READ) {
		if (nr >= file_cnt)
			return 0;
		if (!fat_read_filedata(nr))
			return 0;
		mmc_load_start(cluster2sector(fat_filedata.startcl));
#ifdef SD_TIMEOUT
		if (state >= ERR)
			return 0;
#endif
		filemode = FILEMODE_READ;
		return 1;
	}
#ifndef NO_RECORD
	else if (mode == FILEMODE_WRITE) {
#if defined(DIR_SUPPORT) && !defined(WR_DIR_SUPPORT)
		if (dircl > 0)
			return 0;
#endif
		filemode = FILEMODE_WRITE;
		clcnt = 1;
		startfatoffs = 0;
		startfatcnt = 0;
		fat_filedata.startcl = find_free_cluster();
		return 1;
	}
#endif
	else
		return 0;
}


/** Closes Current File
	@return success
*/
u08up fat_closefile(void)
{
	if (filemode == FILEMODE_READ) {
		mmc_complete_read();
		filemode = FILEMODE_CLOSED;
		return 1;
	}
#ifndef NO_RECORD
	else if (filemode == FILEMODE_WRITE) {
		filemode = FILEMODE_CLOSED;
		return fat_new_file();
	}
#endif
	else
		return 0;
}


/** Finds next Sector for Read
	Works on global var sect
	@return success
*/
u08up fat_getnextsector_rd(void)
{
	u16up startcl = sector2cluster(sect);	// Rechnet den aktuellen Cluster der Datei aus
	u16up fatsect, fatcl;
	u08up offscl;

	if (startcl == sector2cluster(sect+1)) {
		sect++;
		return 1;
	}
	fatsect = fat_pos + (startcl>>8);	// Sektor der FAT mit dem Cluster der Datei
	offscl = (u08)startcl;	// Eintragsposition im FAT-Sektor bestimmen
	mmc_read_sector(fatsect, sharedmem);	// FAT-Sektor lesen
#ifdef SD_TIMEOUT
	if (state >= ERR)
		return 0;
#endif
	fatcl = sharedmem[offscl<<1];	// FAT-Eintrag lesen
	fatcl |= sharedmem[(offscl<<1)+1] << 8;
	if (fatcl >= 2 && fatcl < 0xfff0) {
		sect = cluster2sector(fatcl);
		return 1;
	}
	else {
#ifdef WR_DIR_SUPPORT
		dirfatflag |= NEW_CLUSTER_ALLOC;
#endif
		return 0;
	}
}


#ifndef NO_RECORD

/** Finds next Sector for Write
	Works on global var sect
	@return success
*/
u08up fat_getnextsector_wr(void)
{
	u16up startcl = sector2cluster(sect); // Rechnet den aktuellen Cluster der Datei aus

	if (startcl == sector2cluster(sect+1)) {
		sect++;
		return 1;
	}
	clcnt++;
#ifdef NOSECTORSEARCH
	sect++;
#else
	sect = cluster2sector(find_free_cluster());
	// TODO - Vollerkennung
#endif
	return 1;
}
#endif


#ifdef DIR_SUPPORT

/** Enters the current directory in fat_filedata.
*/
boolup fat_enter_dir(void)
{
	dircl = fat_filedata.startcl;
	fat_count_files();
	return 1;
}


/**
*/
boolup fat_leave_dir(void)
{
	if (dircl == 0)
		return 0;
	fat_read_filedata(DIR_UP_FILE);
	dircl = fat_filedata.startcl;
	fat_count_files();
	return 1;
}

#endif
