/*
 * a.lp_mp3 - Open Source Atmel AVR based MP3 Player
 * Copyright (c) 2003-2004 K. John '2B|!2B' Crispin
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA02111-1307USA
 *
 * Feedback, Bugs.... mail john{AT}phrozen.org
 *
 */


// this code includes a few functions that were originally made for a ti msp340.
// i found it on www.mikrocontroller.net once upon a time
// the ripped functions are
// MMC_get_R1
// MMC_get_R2
// the cs_deselect; spi_io(0xff); CS_select; in the while loop in MMC_init doesn't work without
// unfortunatley i lost the link.
// if you know where this code can be found, let me know so i can put a link here.

#include <avr/io.h>
#include <avr/interrupt.h>

#include "types.h"
#include "mrmidi2.h"
#include "delay.h"
#include "spi.h"
#include "mmc.h"
#ifndef LED_DISP
	#include "lcd.h"
#else
	#define lcd_data(a) ;
	#define lcd_setcur(a, b) ;
	#define lcd_data(a) ;
	#define lcd_string(a, b) ;
	#define lcd_init() ;
	#define lcd_hex_u08(a) ;
	#define lcd_hex_u16(a) ;
	#define lcd_number(a, b) ;
	#define lcd_number_k(a) ;
#endif
#include "fat16.h"

#define MMC_CS_DDR	DDRB
#define MMC_CS_PIN	PINB
#define MMC_CS_PORT	PORTB
#define PIN_MMC_CS PB2

// look in datasheet for complete list of commands
#define MMC_CMD_0_GO_IDLE				0
#define MMC_CMD_1_SEND_OP_COND			1
#define MMC_CMD_9_SEND_CSD				9
#define MMC_CMD_10_SEND_CID				10
#define MMC_CMD_12_STOP					12
#define MMC_CMD_13_SEND_STATUS			13
#define MMC_CMD_16_BLOCKLEN				16
#define MMC_CMD_17_READ_SINGLE			17
#define MMC_CMD_18_READ_MULTIPLE		18
#define MMC_CMD_24_WRITE_SINGLE			0x18

// bit definitions for R1
#define MMC_R1_IN_IDLE 	0x01
#define MMC_R1B_BUSY_BYTE 0x00

// different defines for tokens etc ...
#define MMC_START_TOKEN_SINGLE  0xfe
#define MMC_START_TOKEN_MULTI   0xfc


// starts the read process of a sector
static void MMC_get_sec_start(u32up sector);
// starts the read process of a part of a sector
//static void MMC_get_part_sec_start(u16 sectorh, u16 sectorl, u16 offset, u16 length);
// gets the next byte from the MMC card
static u08up MMC_get_sec_next(void);
// stop read process of a sector
static void MMC_get_sec_stop(void);

// deselcets the MMC
inline static void MMC_CS_deselect(void);
//selects the MMC for transmission
inline static void MMC_CS_select(void);

// sends receives a byte on the spi bus
//static u08 MMC_io(u08 data);

inline static void MMC_cleanup(void);


//---------------------------------------


#ifdef OLD_CARDSCAN
	u32 last_sector;
#endif
#ifdef SD_TIMEOUT
	extern u08 sdtimeout;
#endif
#ifdef MMC_OPEN_CHECK
	u08 mmc_open;
#endif

static u16 current_blocklen = 0;


// sends a block of data from the mem over the spi bus
static void spi_io_mem(u08* data, u16up length) {
	// transmit 'length' bytes over spi
	while (length) {
		spi_io(*data);
		data++;
		length--;
	}
}

// send cmd + arguments + default crc for init command. once in spi mode
// we dont need crc so we can keep this constant
static void MMC_send_cmd(u08up cmd, u32up data) {
	// default command sequence
	u08 buffer[6] ;
	// fill sequence with our specific data
	buffer[0]=0x40 | cmd;
	buffer[1]=data>>24;
	buffer[2]=data>>16;
	buffer[3]=data>>8;
	buffer[4]=data;
	buffer[5]=0x95;
	// dispatch data
	spi_io_mem(buffer,6);
}

// gets a 1 byte long R1
static u08up MMC_get_R1(void) {
	u08up retval;
	u08up max_errors = 128;
	// wait for first valid response byte
	do {
		retval = spi_io(0xff);
		max_errors--;
	}
	while ((retval & 0x80) && (max_errors>0)
#ifdef SD_TIMEOUT
		&& sdtimeout
#endif
	);
	return retval;
}

// waits for the card to send the start data block token
static void MMC_wait_for_start_token(u08up max_errors) {
	u08up retval;
	do {
		// get a byte from the spi bus
		retval = spi_io(0xff);
		// keep track of the trys
		max_errors--;
	}
	while ((retval != MMC_START_TOKEN_SINGLE)
#ifdef SD_TIMEOUT
		&& sdtimeout
#endif
	);
}

// selects the CS for spi xfer
inline static void MMC_CS_select(void) {
	// pull down the MMC CS line
	cbi(MMC_CS_PORT, PIN_MMC_CS);
}

// deselects the CS for spi xfer
inline static void MMC_CS_deselect(void) {
	// pull up the MMC CS card
	sbi(MMC_CS_PORT, PIN_MMC_CS);
}


// stops the MMC transmission and sends the clock cycles needed by the mmc for cleanup
inline static void MMC_cleanup(void) {
	// deselect the MMC card
	MMC_CS_deselect();
	// pulse the SCK 16 times
	spi_io(0xff);
	spi_io(0xff);
}

// gets n bytes plus crc from spi bus
inline static void MMC_get_data(u08* ptr_data, u16up length) {
	MMC_wait_for_start_token(128);
	while (length) {
		*ptr_data = spi_io(0xff);
		//USART_sendint(*ptr_data);
		//USART_send(' ');
		length--;
		ptr_data++;
	}
	// get the 2 CRC bytes
	spi_io(0xff);
	spi_io(0xff);
}


// reads the CSD reg from the card
inline static void MMC_get_CSD(u08 *ptr_data) {
	// select card
	MMC_CS_select();
	// tell the MMC card that we want to know its status
	MMC_send_cmd(MMC_CMD_9_SEND_CSD, 0);
	// get the response
	MMC_get_R1();
	// get the register data
	MMC_get_data(ptr_data, 16);
	// cleanup behind us
	MMC_cleanup();
}

// set the BLOCKLEN for transmissions
inline static void MMC_set_blocklen(u16up blocklen) {
	// make sure this block len is not already set
	if (current_blocklen != blocklen) {
		current_blocklen= blocklen;
		// select card
		MMC_CS_select();
		// tell the MMC card that we want to know its status
		MMC_send_cmd(MMC_CMD_16_BLOCKLEN, blocklen);
		// get the response
		MMC_get_R1();
		// cleanup behind us
		MMC_cleanup();
	}
}



#ifdef OLD_CARDSCAN

// returns the :
// 		size of the card in MB ( ret * 1024^2) == bytes
// 		sector count and multiplier MB are in u08 == C_SIZE / (2^(9-C_SIZE_MULT))
// 		name of the media
static void MMC_get_volume_info(void) {//VOLUME_INFO* vinf) {
	u08 data[16];
	// read the CSD register
	MMC_get_CSD(data);
	// get the C_SIZE value. bits [73:62] of data
	// [73:72] == data[6] && 0x03
	// [71:64] == data[7]
	// [63:62] == data[8] && 0xc0
	last_sector = data[6] & 0x03;
	last_sector <<= 8;
	last_sector += data[7];
	last_sector <<= 2;
	last_sector += (data[8] & 0xc0) >> 6;
}
#endif

// sets up the pins used by the MMC interface
void MMC_hw_init(void) {
	// set MMC CS pin high output
	sbi(MMC_CS_PORT,PIN_MMC_CS);
	sbi(MMC_CS_DDR, PIN_MMC_CS);
}

// starts up the MMC card
u08up MMC_init(void) {
	u08up i;
	u08up res;
	// the data sheet says that the MMC needs 74 clock pulses to startup
	// 10*8== 80; 80>76
	for (i = 0; i < 100; i++) {
		spi_io(0xff);
	}

	// select card
	MMC_CS_select();
	// put MMC in idle
#ifdef SD_TIMEOUT
	sdtimeout = SDTIMEOUT;
#endif
	MMC_send_cmd(MMC_CMD_0_GO_IDLE, 0);
	// get the response
	res = MMC_get_R1();

	//if(res != 0x01) USART_sendstring("Error R1-init",0x01);
	if (res != 0x01) {
#ifdef SD_TIMEOUT
		sdtimeout = 0;
#endif
		return 0;
	}

	// tell the MMC to start its init process by sending the MMC_CMD_1_SEND_OP_COND comand
	// until the response has the idle bit set to 0
	while (res == 0x01) {
		// deselect card
		MMC_CS_deselect();
		// send 32 clock pulses
		spi_io(0xff);
		spi_io(0xff);
		spi_io(0xff);
		spi_io(0xff);
		// select card
		MMC_CS_select();
		// send wake up signal s.t. MMC leaves idle state and switches to operation mode
		MMC_send_cmd(MMC_CMD_1_SEND_OP_COND, 0);
		// get response
		res = MMC_get_R1();
	}
	// cleanup behind us
	MMC_cleanup();

	// now increase MMC card interface speed
	outp(((1<<MSTR)|(1<<SPE) ), SPCR );	// enable SPI interface (8MHz/4=2MHz)
	SPSR &= ~1;	// SPI2X off

	// find out some info on card
#ifdef OLD_CARDSCAN
	MMC_get_volume_info();
#endif

	MMC_set_blocklen(512);
#ifdef SD_TIMEOUT
	sdtimeout = 0;
#endif
#ifdef MMC_OPEN_CHECK
	mmc_open = 0;
#endif
	return 1;
}

// starts the read process of a sector
static void MMC_get_sec_start(u32up sector) {
	spi_io(0xff);
	spi_io(0xff);
	// turn sectors into byte addr
	sector <<= 9;
	// select card
	MMC_CS_select();
#ifdef SD_TIMEOUT
	cli();
	sdtimeout = SDTIMEOUT;
	sei();
#endif
#ifdef MMC_OPEN_CHECK
	mmc_open = 1;
#endif
	// tell the MMC card that we want to know its status
	MMC_send_cmd(MMC_CMD_17_READ_SINGLE, sector);
	// get the response
	if (MMC_get_R1()) {
		MMC_cleanup();
		return;
	}
	// wait till the mmc starts sending data
	MMC_wait_for_start_token(255);
#ifdef SD_TIMEOUT
	cli();
	sdtimeout = 0;
	sei();
#endif
}

// gets the next byte from the MMC card
static inline u08up MMC_get_sec_next(void) {
/*	u08 tmp = spi_io(0xff);
	return tmp;
*/
	SPDR = 0xFF;
	while((SPSR&(1<<SPIF)) == 0);
	return SPDR;
}

// stop read process of a sector
static void MMC_get_sec_stop(void) {
#ifdef MMC_OPEN_CHECK
	if (mmc_open) {
#endif
		// get the 2 CRC bytes
		spi_io(0xff);
		spi_io(0xff);
		// give enough time
		MMC_cleanup();
#ifdef MMC_OPEN_CHECK
		mmc_open = 0;
	}
#endif
}


#ifndef NO_RECORD

// starts the write process of a sector
static void MMC_write_sec_start(u32up sector) {
	spi_io(0xff);
	spi_io(0xff);
	// turn sectors into byte addr
	sector <<= 9;
	// select card
	MMC_CS_select();
#ifdef SD_TIMEOUT
	cli();
	sdtimeout = SDTIMEOUT;
	sei();
#endif
#ifdef MMC_OPEN_CHECK
	mmc_open = 1;
#endif
	// tell the MMC card that we want to know its status
	MMC_send_cmd(MMC_CMD_24_WRITE_SINGLE, sector);
	// get the response
	if (MMC_get_R1()) {
		MMC_cleanup();
		return;
	}
	// send start byte
	spi_io(0xfe);
#ifdef SD_TIMEOUT
	cli();
	sdtimeout = 0;
	sei();
#endif
}

// writes the next byte to the MMC card
static inline void MMC_write_sec_next(u08up b) {
	volatile u08up flush;
	SPDR = b;
	while((SPSR&(1<<SPIF)) == 0);
	flush = SPDR;
}

// stop write process of a sector
static void MMC_get_sec_stop_w(void){
#ifdef MMC_OPEN_CHECK
	if (mmc_open) {
#endif
		// get the 2 CRC bytes
		spi_io(0xff);
		spi_io(0xff);
		// wait while card is busy
#ifdef SD_TIMEOUT
		cli();
		sdtimeout = SDTIMEOUT;
		sei();
#endif
		while (spi_io(0xff) != 0xff 
#ifdef SD_TIMEOUT
			&& sdtimeout
#endif
		);
		// give enough time
		MMC_cleanup();
#ifdef SD_TIMEOUT
		cli();
		sdtimeout = 0;
		sei();
#endif
#ifdef MMC_OPEN_CHECK
		mmc_open = 0;
	}
#endif
}
#endif



//-------------------------------------------------------------
// High Level MMC-Funktionen

u32 sect;
u16 bufpos=0;

extern u08 flags, fat_directaccess;
#ifdef OLD_CARDSCAN
	#define FAT_FLAG 0x20
#endif

void mmc_read_sector(u32up s, u08 *buf) {
	u16up i;
	MMC_get_sec_start(s);
	for (i=0; i<512; i++)
		buf[i] = MMC_get_sec_next();
	MMC_get_sec_stop();
}

// Lesen beginnen
void mmc_load_start(u32up s) {
	sect = s;
	MMC_get_sec_start(s);
	bufpos = 0;
}


// Byte lesen
u08up mmc_fetch_byte(void) {
	if (bufpos == 512) {
		MMC_get_sec_stop();
#ifdef OLD_CARDSCAN
		if (flags & FAT_FLAG && !fat_directaccess)
#else
		if (!fat_directaccess)
#endif
			fat_getnextsector_rd();
		else
			sect++;
		mmc_load_start(sect);
	}
	bufpos++;
	return MMC_get_sec_next();
}


// Lesen, um die 512 voll zu machen
void mmc_complete_read(void) {
	while (bufpos < 512) {
		MMC_get_sec_next();
		bufpos++;
	}
	MMC_get_sec_stop();
}


void mmc_read_block(u08 *buf, u16up cnt) {
	while (cnt--) {
		if (bufpos == 512) {
			MMC_get_sec_stop();
#ifdef OLD_CARDSCAN
			if (flags & FAT_FLAG && !fat_directaccess)
#else
			if (!fat_directaccess)
#endif
				fat_getnextsector_rd();
			else
				sect++;
			mmc_load_start(sect);
		}
		bufpos++;
		*buf++ = MMC_get_sec_next();
	}
}


#ifndef NO_RECORD

void mmc_write_sector(u32up s, u08 *buf) {
	u16up i;
	MMC_write_sec_start(s);
	for (i=0; i<512; i++)
		MMC_write_sec_next(buf[i]);
	MMC_get_sec_stop_w();
}

// Schreiben beginnen
void mmc_write_start(u32up s) {
	sect = s;
	MMC_write_sec_start(s);
	bufpos = 0;
}


// Byte schreiben
void mmc_write_byte(u08up b) {
	if (bufpos == 512) {
		MMC_get_sec_stop_w();
#ifdef OLD_CARDSCAN
		if (flags & FAT_FLAG && !fat_directaccess)
#else
		if (!fat_directaccess)
#endif
			fat_getnextsector_wr();
		else
			sect++;
		mmc_write_start(sect);
	}
	bufpos++;
	MMC_write_sec_next(b);
}


// Schreiben, um die 512 voll zu machen
void mmc_complete_write(void) {
	while (bufpos < 512) {
		MMC_get_sec_next();
		bufpos++;
	}
	MMC_get_sec_stop_w();
}


void mmc_write_block(u08 *buf, u16up cnt) {
	while (cnt--) {
		mmc_write_byte(*buf);
		buf++;
	}
}
#endif


#ifdef OLD_CARDSCAN

// Daten vergleichen
static u08up mmc_compare_string(u08 *buf, u08up len) {
	u08 i;
	for (i=0; i<len; i++) {
		if (mmc_fetch_byte() != buf[i])
			return 0;
	}
	return 1;
}


extern u08 file_cnt;
extern u16 *file_pos;
extern u16 last;

// Mididateien zhlen
void mmc_count_files(u16up start) {
	u32up anz_s;
	u16up s, i;
	u08up cnt=0;	// Sektorskip-Zhler

	last = start;
	lcd_setcur(BOOTFNUM_X, BOOTFNUM_Y);
	lcd_number(file_cnt, 1);	// Anzeige Anzahl gefundene Dateien Start

	for (s=start; s<last_sector; s++) {
		lcd_setcur(BOOTSECT_X, BOOTSECT_Y);
		lcd_hex_u16(s);	// Anzeige Sektornummer in Echtzeit
		mmc_load_start(s);
		if (mmc_compare_string("MThd", 4)) {	// Sektor anschauen: Ist es der Anfang einer Mididatei?
			file_pos[file_cnt] = s;		// Mididatei merken
			file_cnt++;
			lcd_setcur(BOOTFNUM_X, BOOTFNUM_Y);
			lcd_number(file_cnt, 1);	// Anzeige Anzahl gefundene Dateien in Echtzeit
			cnt = 0;
			if (mmc_fetch_byte() == 0xff)	// Ist es eine von MrMidi2 aufgezeichnete Datei?
				for (i=0; i<499; i++)	// Vorspulen bis zum tatschlichen Header
					mmc_fetch_byte();
			else
				for (i=0; i<9; i++)
					mmc_fetch_byte();
			if (mmc_compare_string("MTrk", 4)) {
				anz_s = (u32)mmc_fetch_byte() << 24;	// Lnge der Datei aus dem Trackheader lesen
				anz_s |= (u32)mmc_fetch_byte() << 16;
				anz_s |= (u32)mmc_fetch_byte() << 8;
				anz_s |= (u32)mmc_fetch_byte();
				s += (u16)((u32)(anz_s+8+6+8-1) >> 9);	// Letzten Sektor der Datei berechnen (spart Suchzeit)
				last = s + 1;	// Eventuell freier Sektor
			}
		}
		else if (file_cnt || s>2000 || (((u08)(~PIND)>>2)&0x3f)) {	// Sollte es keine Mididatei sein
			if (cnt == 16) {	// Skipzhler: 16 nicht-Midi-Sektoren fhren zur Endeerkennung
				mmc_complete_read();
				break;
			}
			else
				cnt++;
		}
		mmc_complete_read();
	}
}
#endif

